@@ -1,6 +1,6 @@
 # vim: ft=python fileencoding=utf-8 sts=4 sw=4 et:
 
-# Copyright 2016-2019 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
+# Copyright 2016-2021 Florian Bruhin (The Compiler) <mail@qutebrowser.org>
 #
 # This file is part of qutebrowser.
 #
@@ -15,19 +15,25 @@
 # GNU General Public License for more details.
 #
 # You should have received a copy of the GNU General Public License
-# along with qutebrowser.  If not, see <http://www.gnu.org/licenses/>.
+# along with qutebrowser.  If not, see <https://www.gnu.org/licenses/>.
 
 """Various utilities shared between webpage/webview subclasses."""
 
 import os
+import sys
 import html
+import enum
 import netrc
+import tempfile
+from typing import Callable, Mapping, List, Optional, Iterable, Iterator
 
-from PyQt5.QtCore import QUrl
+from PyQt5.QtCore import QUrl, pyqtBoundSignal
 
 from qutebrowser.config import config
-from qutebrowser.utils import usertypes, message, log, objreg, jinja, utils
+from qutebrowser.utils import (usertypes, message, log, objreg, jinja, utils,
+                               qtutils, version)
 from qutebrowser.mainwindow import mainwindow
+from qutebrowser.misc import guiprocess, objects
 
 
 class CallSuper(Exception):
@@ -45,7 +51,9 @@ def custom_headers(url):
 
     conf_headers = config.instance.get('content.headers.custom', url=url)
     for header, value in conf_headers.items():
-        headers[header.encode('ascii')] = value.encode('ascii')
+        encoded_header = header.encode('ascii')
+        encoded_value = b"" if value is None else value.encode('ascii')
+        headers[encoded_header] = encoded_value
 
     accept_language = config.instance.get('content.headers.accept_language',
                                           url=url)
@@ -74,15 +82,19 @@ def authentication_required(url, authenticator, abort_on):
     return answer
 
 
-def javascript_confirm(url, js_msg, abort_on, *, escape_msg=True):
+def _format_msg(msg: str) -> str:
+    """Convert message to HTML suitable for rendering."""
+    return html.escape(msg).replace('\n', '<br />')
+
+
+def javascript_confirm(url, js_msg, abort_on):
     """Display a javascript confirm prompt."""
     log.js.debug("confirm: {}".format(js_msg))
     if config.val.content.javascript.modal_dialog:
         raise CallSuper
 
-    js_msg = html.escape(js_msg) if escape_msg else js_msg
     msg = 'From <b>{}</b>:<br/>{}'.format(html.escape(url.toDisplayString()),
-                                          js_msg)
+                                          _format_msg(js_msg))
     urlstr = url.toString(QUrl.RemovePassword | QUrl.FullyEncoded)
     ans = message.ask('Javascript confirm', msg,
                       mode=usertypes.PromptMode.yesno,
@@ -90,7 +102,7 @@ def javascript_confirm(url, js_msg, abort_on, *, escape_msg=True):
     return bool(ans)
 
 
-def javascript_prompt(url, js_msg, default, abort_on, *, escape_msg=True):
+def javascript_prompt(url, js_msg, default, abort_on):
     """Display a javascript prompt."""
     log.js.debug("prompt: {}".format(js_msg))
     if config.val.content.javascript.modal_dialog:
@@ -98,9 +110,8 @@ def javascript_prompt(url, js_msg, default, abort_on, *, escape_msg=True):
     if not config.val.content.javascript.prompt:
         return (False, "")
 
-    js_msg = html.escape(js_msg) if escape_msg else js_msg
     msg = '<b>{}</b> asks:<br/>{}'.format(html.escape(url.toDisplayString()),
-                                          js_msg)
+                                          _format_msg(js_msg))
     urlstr = url.toString(QUrl.RemovePassword | QUrl.FullyEncoded)
     answer = message.ask('Javascript prompt', msg,
                          mode=usertypes.PromptMode.text,
@@ -113,7 +124,7 @@ def javascript_prompt(url, js_msg, default, abort_on, *, escape_msg=True):
         return (True, answer)
 
 
-def javascript_alert(url, js_msg, abort_on, *, escape_msg=True):
+def javascript_alert(url, js_msg, abort_on):
     """Display a javascript alert."""
     log.js.debug("alert: {}".format(js_msg))
     if config.val.content.javascript.modal_dialog:
@@ -122,9 +133,8 @@ def javascript_alert(url, js_msg, abort_on, *, escape_msg=True):
     if not config.val.content.javascript.alert:
         return
 
-    js_msg = html.escape(js_msg) if escape_msg else js_msg
     msg = 'From <b>{}</b>:<br/>{}'.format(html.escape(url.toDisplayString()),
-                                          js_msg)
+                                          _format_msg(js_msg))
     urlstr = url.toString(QUrl.RemovePassword | QUrl.FullyEncoded)
     message.ask('Javascript alert', msg, mode=usertypes.PromptMode.alert,
                 abort_on=abort_on, url=urlstr)
@@ -132,7 +142,7 @@ def javascript_alert(url, js_msg, abort_on, *, escape_msg=True):
 
 # Needs to line up with the values allowed for the
 # content.javascript.log setting.
-_JS_LOGMAP = {
+_JS_LOGMAP: Mapping[str, Callable[[str], None]] = {
     'none': lambda arg: None,
     'debug': log.js.debug,
     'info': log.js.info,
@@ -148,54 +158,86 @@ def javascript_log_message(level, source, line, msg):
     logger(logstring)
 
 
-def ignore_certificate_errors(url, errors, abort_on):
+def ignore_certificate_error(
+        *,
+        request_url: QUrl,
+        first_party_url: QUrl,
+        error: usertypes.AbstractCertificateErrorWrapper,
+        abort_on: Iterable[pyqtBoundSignal],
+) -> bool:
     """Display a certificate error question.
 
     Args:
-        url: The URL the errors happened in
-        errors: A list of QSslErrors or QWebEngineCertificateErrors
+        request_url: The URL of the request where the errors happened.
+        first_party_url: The URL of the page we're visiting. Might be an invalid QUrl.
+        error: A single error.
+        abort_on: Signals aborting a question.
 
     Return:
         True if the error should be ignored, False otherwise.
     """
-    ssl_strict = config.instance.get('content.ssl_strict', url=url)
-    log.webview.debug("Certificate errors {!r}, strict {}".format(
-        errors, ssl_strict))
+    conf = config.instance.get('content.tls.certificate_errors', url=request_url)
+    log.network.debug(f"Certificate error {error!r}, config {conf}")
 
-    for error in errors:
-        assert error.is_overridable(), repr(error)
+    assert error.is_overridable(), repr(error)
 
-    if ssl_strict == 'ask':
+    # We get the first party URL with a heuristic - with HTTP -> HTTPS redirects, the
+    # scheme might not match.
+    is_resource = (
+        first_party_url.isValid() and
+        not request_url.matches(
+            first_party_url,
+            QUrl.RemoveScheme))  # type: ignore[arg-type]
+
+    if conf == 'ask' or conf == 'ask-block-thirdparty' and not is_resource:
         err_template = jinja.environment.from_string("""
-            Errors while loading <b>{{url.toDisplayString()}}</b>:<br/>
-            <ul>
-            {% for err in errors %}
-                <li>{{err}}</li>
-            {% endfor %}
-            </ul>
-        """.strip())
-        msg = err_template.render(url=url, errors=errors)
+            {% if is_resource %}
+            <p>
+                Error while loading resource <b>{{request_url.toDisplayString()}}</b><br/>
+                on page <b>{{first_party_url.toDisplayString()}}</b>:
+            </p>
+            {% else %}
+            <p>Error while loading page <b>{{request_url.toDisplayString()}}</b>:</p>
+            {% endif %}
 
-        urlstr = url.toString(QUrl.RemovePassword | QUrl.FullyEncoded)
-        ignore = message.ask(title="Certificate errors - continue?", text=msg,
+            {{error.html()|safe}}
+
+            {% if is_resource %}
+            <p><i>Consider reporting this to the website operator, or set
+            <tt>content.tls.certificate_errors</tt> to <tt>ask-block-thirdparty</tt> to
+            always block invalid resource loads.</i></p>
+            {% endif %}
+
+            Do you want to ignore these errors and continue loading the page <b>insecurely</b>?
+        """.strip())
+        msg = err_template.render(
+            request_url=request_url,
+            first_party_url=first_party_url,
+            is_resource=is_resource,
+            error=error,
+        )
+
+        urlstr = request_url.toString(
+            QUrl.RemovePassword | QUrl.FullyEncoded)  # type: ignore[arg-type]
+        ignore = message.ask(title="Certificate error", text=msg,
                              mode=usertypes.PromptMode.yesno, default=False,
                              abort_on=abort_on, url=urlstr)
         if ignore is None:
             # prompt aborted
             ignore = False
         return ignore
-    elif ssl_strict is False:
-        log.webview.debug("ssl_strict is False, only warning about errors")
-        for err in errors:
-            # FIXME we might want to use warn here (non-fatal error)
-            # https://github.com/qutebrowser/qutebrowser/issues/114
-            message.error('Certificate error: {}'.format(err))
+    elif conf == 'load-insecurely':
+        message.error(f'Certificate error: {error}')
         return True
-    elif ssl_strict is True:
+    elif conf == 'block':
         return False
-    else:
-        raise ValueError("Invalid ssl_strict value {!r}".format(ssl_strict))
-    raise utils.Unreachable
+    elif conf == 'ask-block-thirdparty' and is_resource:
+        log.network.error(
+            f"Certificate error in resource load: {error}\n"
+            f"  request URL:     {request_url.toDisplayString()}\n"
+            f"  first party URL: {first_party_url.toDisplayString()}")
+        return False
+    raise utils.Unreachable(conf, is_resource)
 
 
 def feature_permission(url, option, msg, yes_action, no_action, abort_on,
@@ -257,10 +299,8 @@ def get_tab(win_id, target):
         target: A usertypes.ClickTarget
     """
     if target == usertypes.ClickTarget.tab:
-        win_id = win_id
         bg_tab = False
     elif target == usertypes.ClickTarget.tab_bg:
-        win_id = win_id
         bg_tab = True
     elif target == usertypes.ClickTarget.window:
         tabbed_browser = objreg.get('tabbed-browser', scope='window',
@@ -286,10 +326,24 @@ def get_user_stylesheet(searching=False):
         with open(filename, 'r', encoding='utf-8') as f:
             css += f.read()
 
-    if (config.val.scrolling.bar == 'never' or
-            config.val.scrolling.bar == 'when-searching' and not searching):
+    setting = config.val.scrolling.bar
+    if setting == 'overlay' and utils.is_mac:
+        setting = 'when-searching'
+
+    if setting == 'never' or setting == 'when-searching' and not searching:
         css += '\nhtml > ::-webkit-scrollbar { width: 0px; height: 0px; }'
 
+    if (objects.backend == usertypes.Backend.QtWebEngine and
+            version.qtwebengine_versions().chromium_major in [69, 73, 80, 87] and
+            config.val.colors.webpage.darkmode.enabled and
+            config.val.colors.webpage.darkmode.policy.images == 'smart' and
+            config.val.content.site_specific_quirks.enabled and
+            'misc-mathml-darkmode' not in config.val.content.site_specific_quirks.skip):
+        # WORKAROUND for MathML-output on Wikipedia being black on black.
+        # See https://bugs.chromium.org/p/chromium/issues/detail?id=1126606
+        css += ('\nimg.mwe-math-fallback-image-inline, '
+                'img.mwe-math-fallback-image-display { filter: invert(100%); }')
+
     return css
 
 
@@ -340,3 +394,127 @@ def netrc_authentication(url, authenticator):
     authenticator.setPassword(password)
 
     return True
+
+
+class FileSelectionMode(enum.Enum):
+    """Mode to use for file selectors in choose_file."""
+
+    single_file = enum.auto()
+    multiple_files = enum.auto()
+    folder = enum.auto()
+
+
+def choose_file(qb_mode: FileSelectionMode) -> List[str]:
+    """Select file(s)/folder for uploading, using external command defined in config.
+
+    Args:
+        qb_mode: File selection mode
+
+    Return:
+        A list of selected file paths, or empty list if no file is selected.
+        If multiple is False, the return value will have at most 1 item.
+    """
+    command = {
+        FileSelectionMode.single_file: config.val.fileselect.single_file.command,
+        FileSelectionMode.multiple_files: config.val.fileselect.multiple_files.command,
+        FileSelectionMode.folder: config.val.fileselect.folder.command,
+    }[qb_mode]
+    use_tmp_file = any('{}' in arg for arg in command[1:])
+    if use_tmp_file:
+        with tempfile.NamedTemporaryFile(
+            prefix='qutebrowser-fileselect-',
+            delete=False,
+        ) as handle:
+            tmpfilename = handle.name
+        with utils.cleanup_file(tmpfilename):
+            command = (
+                command[:1] +
+                [arg.replace('{}', tmpfilename) for arg in command[1:]]
+            )
+            return _execute_fileselect_command(
+                command=command,
+                qb_mode=qb_mode,
+                tmpfilename=tmpfilename,
+            )
+    else:
+        return _execute_fileselect_command(
+            command=command,
+            qb_mode=qb_mode,
+        )
+
+
+def _execute_fileselect_command(
+    command: List[str],
+    qb_mode: FileSelectionMode,
+    tmpfilename: Optional[str] = None
+) -> List[str]:
+    """Execute external command to choose file.
+
+    Args:
+        multiple: Should selecting multiple files be allowed.
+        tmpfilename: Path to the temporary file if used, otherwise None.
+
+    Return:
+        A list of selected file paths, or empty list if no file is selected.
+        If multiple is False, the return value will have at most 1 item.
+    """
+    proc = guiprocess.GUIProcess(what='choose-file')
+    proc.start(command[0], command[1:])
+
+    loop = qtutils.EventLoop()
+    proc.finished.connect(lambda _code, _status: loop.exit())
+    loop.exec()
+
+    if tmpfilename is None:
+        selected_files = proc.stdout.splitlines()
+    else:
+        try:
+            with open(tmpfilename, mode='r', encoding=sys.getfilesystemencoding()) as f:
+                selected_files = f.read().splitlines()
+        except OSError as e:
+            message.error(f"Failed to open tempfile {tmpfilename} ({e})!")
+            selected_files = []
+
+    return list(_validated_selected_files(
+        qb_mode=qb_mode, selected_files=selected_files))
+
+
+def _validated_selected_files(
+    qb_mode: FileSelectionMode,
+    selected_files: List[str],
+) -> Iterator[str]:
+    """Validates selected files if they are.
+
+        * Of correct type
+        * Of correct number
+        * Existent
+
+    Args:
+        qb_mode: File selection mode used
+        selected_files: files selected
+
+    Return:
+        List of selected files that pass the checks.
+    """
+    if qb_mode != FileSelectionMode.multiple_files:
+        if len(selected_files) > 1:
+            message.warning("More than one file/folder chosen, using only the first")
+            selected_files = selected_files[:1]
+    for selected_file in selected_files:
+        if not os.path.exists(selected_file):
+            message.warning(f"Ignoring non-existent file '{selected_file}'")
+            continue
+        if qb_mode == FileSelectionMode.folder:
+            if not os.path.isdir(selected_file):
+                message.warning(
+                    f"Expected folder but got file, ignoring '{selected_file}'"
+                )
+                continue
+        else:
+            # pylint: disable=else-if-used
+            if not os.path.isfile(selected_file):
+                message.warning(
+                    f"Expected file but got folder, ignoring '{selected_file}'"
+                )
+                continue
+        yield selected_file
